from collections import OrderedDict


class SelectionHandler:
    def __init__(self, hp_config):
        self.hp_config = hp_config
        self.combinations = hp_config.get_combinations()
        self.defaults = hp_config.get_defaults()
        self.current_options = OrderedDict()
        self.selected_params = OrderedDict()
        self.filtered_combinations = []

    def initialize(self):
        self.filtered_combinations = self.combinations.copy()
        self._run_process(current_param_name=None)

    def _run_process(self, current_param_name=None):
        param_name = self._get_next_param_name(current_param_name)
        while param_name is not None:
            self._process_param(param_name)
            param_name = self._get_next_param_name(param_name)

    def _get_next_param_name(self, current_param_name=None):
        seen = False
        for combination in self.filtered_combinations:
            for param_name in combination:
                if current_param_name is None:
                    return param_name
                elif param_name == current_param_name:
                    seen = True
                elif seen:
                    return param_name
            seen = False
        return None

    def _process_param(self, param_name):
        self.current_options[param_name] = self._collect_options(param_name, self.filtered_combinations)
        self.selected_params[param_name] = self._select_value(
            param_name, self.defaults.get(param_name, None), self.current_options[param_name]
        )
        filtered_combinations = self._filter_combinations(
            param_name, self.filtered_combinations, self.selected_params[param_name]
        )
        self.filtered_combinations = self._remove_dups(filtered_combinations)

    def _remove_dups(self, filtered_combinations):
        unique_combinations = []
        seen = set()

        for combination in filtered_combinations:
            combination_hash = self._hash_combination(combination)
            if combination_hash not in seen:
                seen.add(combination_hash)
                unique_combinations.append(combination)

        return unique_combinations

    def _hash_combination(self, combination):
        hash_parts = []
        for key, value in combination.items():
            if isinstance(value, dict):
                hash_parts.append(f"{key}:{self._hash_combination(value)}")
            elif isinstance(value, list):
                hash_parts.append(f"{key}:{tuple(sorted(value))}")
            else:
                hash_parts.append(f"{key}:{value}")
        return tuple(sorted(hash_parts))

    def _process_nested_dict(self, param_name):
        for sub_param in self.combinations[param_name]:
            full_param_name = f"{param_name}.{sub_param}"
            self._process_param(full_param_name)

    def _collect_options(self, param_name, combinations):
        if isinstance(combinations[0][param_name], dict):
            sub_combinations = []
            for combination in combinations:
                sub_combinations.append(combination[param_name])

            comb_dct = OrderedDict()
            for combination in sub_combinations:
                for key in combination.keys():
                    comb_dct[key] = self._collect_options(key, sub_combinations)
            return comb_dct

        options_set = set()
        for combination in combinations:
            param = combination[param_name]
            if isinstance(param, list):
                options_set.update(set(param))
            else:
                options_set.add(param)
        return options_set

    def _select_value(self, param_name, param_defaults, current_param_options):
        if isinstance(current_param_options, dict):
            selected_values = OrderedDict()
            for key in current_param_options.keys():
                selected_values[key] = self._select_value(
                    key, param_defaults.get(key, None) if param_defaults else None, current_param_options[key]
                )
            return selected_values
        if param_defaults:
            for default in reversed(param_defaults):
                if isinstance(default, list):
                    if set(default).issubset(current_param_options):
                        return default
                elif default in current_param_options:
                    return default
        # if no default is found, set the first option as default
        return next(iter(current_param_options))

    def _filter_combinations(self, param_name, current_combinations, selected_param_value):
        if isinstance(selected_param_value, dict):
            sub_combinations = []
            for combination in current_combinations:
                append_flag = True
                for key in selected_param_value.keys():
                    result = self._filter_combinations(key, [combination[param_name]], selected_param_value[key])
                    if len(result) == 0:  # if the selected params are not in "combination", filter it out
                        append_flag = False
                        break
                if append_flag:
                    sub_combinations.append(combination)
            return sub_combinations
        else:
            filtered_combinations = [
                comb for comb in current_combinations if self._compare_values(comb[param_name], selected_param_value)
            ]
            return filtered_combinations

    def _compare_values(self, value1, value2):
        if isinstance(value1, list) and isinstance(value2, list):
            return set(value1) == set(value2)
        else:
            return value1 == value2

    def update_param(self, param_name, selected_value):
        self.defaults[param_name] = self._update_defaults(
            param_name, self.defaults.get(param_name, None), selected_value
        )
        self.current_options = OrderedDict()
        self.selected_params = OrderedDict()
        self.filtered_combinations = self.combinations.copy()
        self._run_process()

    def _update_defaults(self, param_name, param_defaults, selected_value):
        if isinstance(selected_value, dict):
            if param_defaults is None:
                defaults = {}
            else:
                defaults = param_defaults.copy()

            for key in selected_value.keys():
                defaults[key] = self._update_defaults(key, param_defaults.get(key, None), selected_value[key])
            return defaults

        if param_defaults is None:
            defaults = []
        else:
            defaults = param_defaults.copy()

        defaults.append(selected_value)
        return defaults

    def get_current_state(self):
        return {
            "selected_params": self.selected_params,
            "current_options": self.current_options,
            "filtered_combinations": self.filtered_combinations,
        }



import ast
import logging
import textwrap
from typing import Optional, Tuple

from .hp import HP

#logging.basicConfig(level=logging.WARNING, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


# TODO: consider moving these functions to ast_analyzer
def get_hp_function_node(tree: ast.Module) -> Optional[ast.FunctionDef]:
    """
    Finds the first function with 'hp' in its signature in the given abstract syntax tree.

    :param tree: The abstract syntax tree to search in
    :return: The function definition node with 'hp' in its signature
    :raises ValueError: If no function with 'hp' is found, or if multiple functions \
    with 'hp' are found or if there's a function with 'hp' and another argument
    """
    hp_functions = []

    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef):
            has_hp = False
            has_other_args = False
            for arg in node.args.args:
                if arg.arg == "hp":
                    has_hp = True
                elif arg.arg != "self":  # Ignore 'self' for methods
                    has_other_args = True

            if has_hp:
                if has_other_args:
                    raise ValueError(f"Function '{node.name}' has 'hp' and other arguments in its signature.")
                hp_functions.append(node)

    if len(hp_functions) > 1:
        raise ValueError("Multiple functions with 'hp' in their signatures found.")
    elif len(hp_functions) == 0:
        raise ValueError("No function with 'hp' in its signature found.")

    return hp_functions[0]


def find_hp_function_body_and_name(source_code: str) -> Optional[Tuple[str, str]]:
    dedented_source = textwrap.dedent(source_code)
    tree = ast.parse(dedented_source)

    function_node = get_hp_function_node(tree)

    if function_node is None:
        return None

    # Ensure that both function_name and function_body are strings
    function_name = function_node.name if isinstance(function_node.name, str) else None
    function_body = ast.get_source_segment(dedented_source, function_node)

    # If function_body or function_name is None, return None
    if function_name is None or function_body is None:
        return None

    return function_name, function_body


def remove_function_signature(source: str) -> str:
    lines = source.split("\n")
    for i, line in enumerate(lines):
        if line.strip().endswith(":"):
            return "\n".join(lines[i + 1 :])
    raise ValueError("Could not find function signature")


def query_combinations(combinations, query):
    """
    Filter combinations based on the provided query.

    Args:
    combinations (list): List of dictionaries, each representing a combination of hyperparameters.
    query (dict): Dictionary of key-value pairs to filter the combinations.

    Returns:
    list: Filtered list of combinations that match the query criteria.
    """

    def match_combination(combination, query):
        for key, value in query.items():
            if key not in combination or combination[key] != value:
                return False
        return True

    return [comb for comb in combinations if match_combination(comb, query)]



import ast

# from .logging_utils import configure_logging
#logger = configure_logging()
from logging import getLogger
from typing import Any, Dict, List, Optional, Set, Tuple

logger = getLogger(__name__)


class VariableGraphBuilder(ast.NodeVisitor):
    def __init__(self):
        self.graph = {}
        self.current_path = []

    def visit_Assign(self, node):
        target = self.get_target_name(node.targets[0])
        logger.debug(f"VariableGraphBuilder: Visiting assignment to {target}")
        self.current_path = [target]
        self.graph[target] = self.build_subgraph(node.value)

    def build_subgraph(self, node):
        logger.debug(f"VariableGraphBuilder: Building subgraph for node type {type(node).__name__}")
        if isinstance(node, ast.Dict):
            return {self.get_key_name(k): self.build_subgraph(v) for k, v in zip(node.keys, node.values)}
        elif isinstance(node, ast.Call):
            return {
                "__call__": self.get_target_name(node.func),
                "args": [self.get_node_value(arg) for arg in node.args],
                "kwargs": {kw.arg: self.get_node_value(kw.value) for kw in node.keywords},
            }
        elif isinstance(node, (ast.Name, ast.Attribute)):
            return self.get_target_name(node)
        elif isinstance(node, ast.Constant):
            return node.value
        else:
            return {"__unknown__": ast.unparse(node)}

    def get_node_value(self, node):
        if isinstance(node, ast.Constant):
            return node.value
        elif isinstance(node, (ast.Name, ast.Attribute)):
            return self.get_target_name(node)
        else:
            return ast.unparse(node)

    def get_target_name(self, node):
        if isinstance(node, ast.Name):
            return node.id
        elif isinstance(node, ast.Attribute):
            return f"{self.get_target_name(node.value)}.{node.attr}"
        else:
            return "Unknown"

    def get_key_name(self, node):
        if isinstance(node, ast.Str):
            return node.s
        return self.get_target_name(node)


class HPCall:
    def __init__(
        self,
        lineno: int,
        col_offset: int,
        method_name: str,
        implicit_name: Optional[str] = None,
        has_explicit_name: Optional[bool] = None,
    ):
        self.lineno = lineno
        self.col_offset = col_offset
        self.method_name = method_name
        self.implicit_name = implicit_name
        self.has_explicit_name = has_explicit_name
        self.call_index = 0  # Will be set later


class HPCallAnalyzer(ast.NodeVisitor):
    def __init__(self, variable_graph):
        self.variable_graph = variable_graph
        self.results = {}
        self.current_assignment = None
        self.current_complex_assignment = None
        self.hp_calls = []

    def visit_Assign(self, node):
        logger.debug(f"HPCallAnalyzer: Visiting assignment on line {node.lineno}")
        self.current_assignment = node
        start_line = node.lineno
        end_line = self.get_last_line(node)

        if isinstance(node.value, (ast.Dict, ast.Call)):
            self.current_complex_assignment = node

        self.results[start_line] = {
            "code": ast.unparse(node),
            "hp_calls": [],
            "end_line": end_line,
        }
        self.generic_visit(node)

        self.current_complex_assignment = None
        self.current_assignment = None

    def visit_Call(self, node):
        if self.is_hp_call(node):
            logger.debug(f"HPCallAnalyzer: hp call detected on line {node.lineno}")
            line_number = node.lineno
            call_index = len([c for c in self.hp_calls if c.lineno == line_number])

            method_name = node.func.attr
            inferred_name = self.infer_name(node)

            if not node.has_explicit_name and inferred_name and self.is_valid_inference(inferred_name):
                implicit_name = inferred_name
            else:
                implicit_name = None

            hp_call = HPCall(
                line_number,
                call_index,
                method_name,
                node.has_explicit_name,
                implicit_name,
            )
            self.hp_calls.append(hp_call)

            if (
                self.current_complex_assignment
                and self.current_complex_assignment.lineno
                <= line_number
                <= self.get_last_line(self.current_complex_assignment)
            ):
                logger.debug(
                    f"HPCallAnalyzer: hp call is part of a complex\
                          assignment starting on line {self.current_complex_assignment.lineno}"
                )
                line_number = self.current_complex_assignment.lineno

            if line_number in self.results:
                self.results[line_number]["hp_calls"].append(hp_call)
            else:
                self.results[line_number] = {
                    "code": ast.unparse(self.current_complex_assignment or node),
                    "hp_calls": [hp_call],
                    "end_line": self.get_last_line(self.current_complex_assignment or node),
                }

        self.generic_visit(node)

    def is_valid_inference(self, inferred_name):
        logger.debug(f"HPCallAnalyzer: Checking validity of inferred name: {inferred_name}")
        parts = inferred_name.split(".")
        valid = len(parts) <= 4 and all(not part.startswith("arg") for part in parts)
        logger.debug(f"HPCallAnalyzer: Inferred name {'is' if valid else 'is not'} valid")
        return valid

    def get_last_line(self, node):
        return max(getattr(node, "lineno", 0), getattr(node, "end_lineno", 0))

    def is_hp_call(self, node):
        return (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == "hp"
        )

    # def get_hp_call_name(self, node):
    #     logger.debug(f"HPCallAnalyzer: Attempting to get hp call name for node on line {node.lineno}")
    #     if len(node.args) >= 2:
    #         name = self.get_node_value(node.args[1])
    #         logger.debug(f"HPCallAnalyzer: Found name in second argument: {name}")
    #         return name
    #     for keyword in node.keywords:
    #         if keyword.arg == "name":
    #             name = self.get_node_value(keyword.value)
    #             logger.debug(f"HPCallAnalyzer: Found name in keyword argument: {name}")
    #             return name
    #     logger.debug("HPCallAnalyzer: No explicit name found for hp call")
    #     return None

    def get_node_value(self, node):
        if isinstance(node, ast.Constant):
            return node.value
        elif isinstance(node, ast.Name):
            return node.id
        else:
            return ast.unparse(node)

    def infer_name(self, node):
        logger.debug(f"HPCallAnalyzer: Attempting to infer name for node on line {node.lineno}")
        if self.current_assignment:
            target = self.get_target_name(self.current_assignment.targets[0])
            if isinstance(self.current_assignment.value, ast.Call) and self.is_hp_call(self.current_assignment.value):
                logger.debug(f"HPCallAnalyzer: Direct assignment to hp call detected: {target}")
                return target
            inferred = self.find_node_in_assignment(self.current_assignment.value, node, [target])
            logger.debug(f"HPCallAnalyzer: Inferred name from current assignment: {inferred}")
            return inferred

        for var, subgraph in self.variable_graph.items():
            path = self.find_node_in_graph(subgraph, node)
            if path:
                inferred = ".".join([var] + [p for p in path if p != "kwargs"])
                logger.debug(f"HPCallAnalyzer: Inferred name from variable graph: {inferred}")
                return inferred
        logger.debug("HPCallAnalyzer: Unable to infer name")
        return None

    def find_node_in_assignment(self, value_node, target_node, path):
        logger.debug(f"HPCallAnalyzer: Searching for node in assignment, current path: {'.'.join(path)}")
        if isinstance(value_node, ast.Dict):
            for key, value in zip(value_node.keys, value_node.values):
                if value == target_node:
                    result = ".".join(path + [self.get_node_value(key)])
                    logger.debug(f"HPCallAnalyzer: Found node in dictionary: {result}")
                    return result
                result = self.find_node_in_assignment(value, target_node, path + [self.get_node_value(key)])
                if result:
                    return result
        elif isinstance(value_node, ast.Call):
            for idx, arg in enumerate(value_node.args):
                if arg == target_node:
                    result = ".".join(path + [f"arg{idx}"])
                    logger.debug(f"HPCallAnalyzer: Found node in function argument: {result}")
                    return result
                result = self.find_node_in_assignment(arg, target_node, path + [f"arg{idx}"])
                if result:
                    return result
            for keyword in value_node.keywords:
                if keyword.value == target_node:
                    result = ".".join(path + [keyword.arg])
                    logger.debug(f"HPCallAnalyzer: Found node in keyword argument: {result}")
                    return result
                result = self.find_node_in_assignment(keyword.value, target_node, path + [keyword.arg])
                if result:
                    return result
        logger.debug("HPCallAnalyzer: Node not found in assignment")
        return None

    def find_node_in_graph(self, graph, node, path=None):
        if path is None:
            path = []

        logger.debug(f"HPCallAnalyzer: Searching for node in graph, current path: {'.'.join(path)}")
        if isinstance(graph, dict):
            if (
                graph.get("__call__", "").startswith("hp.")
                and graph["args"]
                and self.get_node_value(node.args[0]) == graph["args"][0]
            ):
                logger.debug(f"HPCallAnalyzer: Found matching hp call in graph at path: {'.'.join(path)}")
                return path
            for key, value in graph.items():
                new_path = self.find_node_in_graph(value, node, path + [key])
                if new_path:
                    return new_path
        elif isinstance(graph, list):
            for i, item in enumerate(graph):
                new_path = self.find_node_in_graph(item, node, path + [str(i)])
                if new_path:
                    return new_path
        logger.debug("HPCallAnalyzer: Node not found in graph")
        return None

    def get_target_name(self, node):
        if isinstance(node, ast.Name):
            return node.id
        elif isinstance(node, ast.Attribute):
            return f"{self.get_target_name(node.value)}.{node.attr}"
        else:
            return "Unknown"


def analyze_hp_calls(code: str) -> Tuple[Dict[int, Dict[str, Any]], List[HPCall]]:
    logger.info("Starting HP calls analysis")
    tree = ast.parse(code)

    logger.debug("Building variable graph")
    graph_builder = VariableGraphBuilder()
    graph_builder.visit(tree)

    logger.debug("Analyzing HP calls")
    analyzer = HPCallAnalyzer(graph_builder.graph)
    analyzer.visit(tree)

    logger.info("HP calls analysis complete")
    return analyzer.results, analyzer.hp_calls


class HPCallVisitor(ast.NodeVisitor):
    def __init__(self):
        self.hp_calls = []
        self.call_index = {}
        self.current_assignment = None
        self.current_path = []

    def visit_Call(self, node):
        if (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == "hp"
        ):
            lineno = node.lineno
            method_name = node.func.attr

            if lineno not in self.call_index:
                self.call_index[lineno] = {}
            if method_name not in self.call_index[lineno]:
                self.call_index[lineno][method_name] = 0
            else:
                self.call_index[lineno][method_name] += 1

            implicit_name = self.get_implicit_name(node)
            has_explicit_name = self.has_explicit_name(node)

            hp_call = HPCall(
                lineno=lineno,
                col_offset=node.col_offset,
                method_name=method_name,
                implicit_name=implicit_name,
                has_explicit_name=has_explicit_name,
            )
            hp_call.call_index = self.call_index[lineno][method_name]
            self.hp_calls.append(hp_call)

        self.generic_visit(node)

    def visit_Assign(self, node):
        self.current_assignment = node  # Track the current assignment
        self.generic_visit(node)
        self.current_assignment = None  # Reset after processing

    def get_implicit_name(self, node):
        if self.current_assignment:
            target = self.get_target_name(self.current_assignment.targets[0])
            if isinstance(self.current_assignment.value, ast.Call) and self.is_hp_call(self.current_assignment.value):
                return target
            return self.find_node_in_assignment(self.current_assignment.value, node, [target])
        return self.find_node_in_current_path(node)

    def has_explicit_name(self, node):
        for kw in node.keywords:
            if kw.arg == "name":
                return True
        return False

    def get_target_name(self, node):
        if isinstance(node, ast.Name):
            return node.id
        elif isinstance(node, ast.Attribute):
            return f"{self.get_target_name(node.value)}.{node.attr}"
        return "Unknown"

    def is_hp_call(self, node):
        return (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == "hp"
        )

    def find_node_in_assignment(self, value_node, target_node, path):
        if isinstance(value_node, ast.Dict):
            for key, value in zip(value_node.keys, value_node.values):
                if value == target_node:
                    return ".".join(path + [self.get_node_value(key)])
                result = self.find_node_in_assignment(value, target_node, path + [self.get_node_value(key)])
                if result:
                    return result
        elif isinstance(value_node, ast.Call):
            for idx, arg in enumerate(value_node.args):
                if arg == target_node:
                    return ".".join(path + [f"arg{idx}"])
                result = self.find_node_in_assignment(arg, target_node, path + [f"arg{idx}"])
                if result:
                    return result
            for keyword in value_node.keywords:
                if keyword.value == target_node:
                    return ".".join(path + [keyword.arg])
                result = self.find_node_in_assignment(keyword.value, target_node, path + [keyword.arg])
                if result:
                    return result
        return None

    def find_node_in_current_path(self, node):
        return ".".join(self.current_path) if self.current_path else None

    def get_node_value(self, node):
        if isinstance(node, ast.Constant):
            return str(node.value)
        elif isinstance(node, ast.Name):
            return node.id
        return ast.unparse(node)


class VariableReferenceCollector(ast.NodeVisitor):
    def __init__(self):
        self.referenced_vars = set()

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Load):
            self.referenced_vars.add(node.id)
        self.generic_visit(node)


def collect_hp_calls(code: str) -> List[HPCall]:
    tree = ast.parse(code)
    visitor = HPCallVisitor()
    visitor.visit(tree)
    return visitor.hp_calls


def find_referenced_vars(code: str) -> Set[str]:
    tree = ast.parse(code)
    collector = VariableReferenceCollector()
    collector.visit(tree)
    return collector.referenced_vars


def inject_names_to_source_code(code: str, hp_calls: List[HPCall]) -> str:
    tree = ast.parse(code)

    class NameInjector(ast.NodeTransformer):
        def __init__(self, hp_calls):
            self.hp_calls = hp_calls
            self.call_index = 0

        def visit_Call(self, node):
            if (
                isinstance(node.func, ast.Attribute)
                and isinstance(node.func.value, ast.Name)
                and node.func.value.id == "hp"
            ):
                if (
                    self.call_index < len(self.hp_calls)
                    and self.hp_calls[self.call_index].implicit_name
                    and self.hp_calls[self.call_index].has_explicit_name is False
                ):
                    node.keywords.append(
                        ast.keyword(arg="name", value=ast.Constant(value=self.hp_calls[self.call_index].implicit_name))
                    )
                self.call_index += 1
            return self.generic_visit(node)

    injector = NameInjector(hp_calls)
    modified_tree = injector.visit(tree)
    return ast.unparse(modified_tree)

def find_independent_select_calls(referenced_vars, hp_calls) -> List[str]:
    independent_vars = {
        call.implicit_name.split(".")[0]
        for call in hp_calls
        if call.implicit_name
        and call.method_name == "select"
        and call.implicit_name.split(".")[0] not in referenced_vars
    }

    return list(independent_vars)


from .config import config
from .core import Hypster, load, save
from .hp import HP
#from .logging_utils import configure_logging
from .selection_handler import SelectionHandler
from .utils import query_combinations

__all__ = ["config", "save", "load", "HP", "SelectionHandler"]



import collections
import itertools
import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union

if TYPE_CHECKING:
    from .hp import HP  # Adjust the import path as needed

logger = logging.getLogger(__name__)


class HPCall(ABC):
    def __init__(
        self, hp_instance: "HP", name: Optional[str] = None, default: Any = None
    ):
        if name is None:
            raise ValueError(
                "`name` argument is missing and must be provided explicitly."
            )
        self.hp = hp_instance
        self.name = self._get_full_name(name)
        self.default = default
        # TODO: improve this
        if default is not None:
            if self.name not in self.hp.defaults:
                self.hp.defaults[self.name] = [default]
            elif default not in self.hp.defaults[self.name]:
                self.hp.defaults[self.name].append(default)
        self.options = None

    def _get_full_name(self, name: str) -> str:
        if self.hp.name_prefix:
            return f"{self.hp.name_prefix}.{name}"
        return name

    def validate_and_process_options(
        self, options: Optional[Union[Dict[str, Any], List[Any]]]
    ) -> Optional[Dict[Any, Any]]:
        if options is not None:
            self._check_options_exists(options)
            if isinstance(options, dict):
                self._validate_dict_keys(options)
            elif isinstance(options, list):
                self._validate_list_values(options)
                options = {v: v for v in options}
            else:
                raise ValueError("Options must be a dictionary or a list.")

            if self.default is not None:
                self._validate_default(self.default, options)
            self.options = options
            return options
        return None

    def _check_options_exists(self, options: Union[Dict[str, Any], List[Any]]):
        if not isinstance(options, (list, dict)) or len(options) == 0:
            raise ValueError("Options must be a non-empty list or dictionary.")

    def _validate_dict_keys(self, options: Dict[str, Any]):
        if not all(isinstance(k, (str, int, bool, float)) for k in options.keys()):
            bad_keys = [
                key
                for key in options.keys()
                if not isinstance(key, (str, int, bool, float))
            ]
            raise ValueError(
                f"Dictionary keys must be str, int, bool, float. Got {bad_keys} instead."
            )

    def _validate_list_values(self, options: List[Any]):
        if not all(isinstance(v, (str, int, bool, float)) for v in options):
            raise ValueError(
                "List values must be one of: str, int, bool, float. For complex types - use a dictionary instead"
            )

    def _validate_default(self, default: Any, options: Dict[str, Any]):
        if default is not None and isinstance(default, list):
            for key in default:
                if key not in options:
                    raise ValueError("Default values must be one of the options.")
        elif (
            default is not None and default not in options
        ):  # TODO: add var/param names to error
            raise ValueError("Default value must be one of the options.")

    def handle_overrides_selections(self) -> Any:
        if self.name in self.hp.overrides:
            result = self._get_result_from_override()
        elif self.name in self.hp.selections:
            result = self._get_result_from_selection()
        elif self.default is not None:
            if isinstance(self.default, list):
                result = []
                for key in self.default:
                    if self.options:
                        if key in self.options:
                            result.append(self.options[key])
                        else:  # TODO: fix
                            raise ValueError("Invalid default not in options.")
                    else:
                        result.append(key)
                snapshot = self.default
            else:
                result = self.options[self.default] if self.options else self.default
                snapshot = self.default
            self.hp.snapshot[self.name] = snapshot
        else:
            raise ValueError(
                f"`{self.name}` has no selections, overrides or defaults provided."
            )

        return result

    def _get_result_from_override(self):
        override_value = self.hp.overrides[self.name]
        logger.debug("Found override for %s: %s", self.name, override_value)
        snapshot = None
        if isinstance(override_value, list):
            result = []
            snapshot = []
            for value in override_value:
                snapshot.append(value)
                if self.options and value in self.options:
                    result.append(self.options[value])
                else:
                    result.append(value)
        elif self.options and override_value in self.options:
            result = self.options[override_value]
        else:
            result = override_value

        if snapshot is None:
            snapshot = override_value
        self.hp.snapshot[self.name] = snapshot
        logger.info("Applied override for %s: %s", self.name, result)
        return result

    def _get_result_from_selection(self):
        # TODO: add error if there's a selection for something that only works with overrides,
        # like text, number input (including multi)
        selected_value = self.hp.selections[self.name]
        logger.debug("Found selection for %s: %s", self.name, selected_value)
        snapshot = None
        if isinstance(selected_value, list):
            result = []
            snapshot = []
            for key in selected_value:
                if self.options and key in self.options:
                    result.append(self.options[key])
                    snapshot.append(key)
                else:
                    raise ValueError(
                        f"Invalid selection '{key}' for '{self.name}'. Not in options: "
                        f"{list(self.options.keys() if self.options else [])}"
                    )
        elif self.options and selected_value in self.options:
            result = self.options[selected_value]
            logger.info("Applied selection for %s: %s", self.name, result)
        else:
            raise ValueError(
                f"Invalid selection '{selected_value}' for '{self.name}'. "
                f"Not in options: {list(self.options.keys() if self.options else [])}"
            )

        if snapshot is None:  # not a list
            snapshot = selected_value
        if self.name not in self.hp.snapshot:
            self.hp.snapshot[self.name] = snapshot
        return result

    # TODO: handle different cases where options is a list, one item, or empty.

    @abstractmethod
    def explore(self) -> Any:
        pass

    @abstractmethod
    def explore(self) -> bool:
        pass


class SelectCall(HPCall):
    def execute(
        self, options: Optional[Union[Dict[str, Any], List[Any]]] = None
    ) -> Any:
        self.options = self.validate_and_process_options(options)
        if self.hp.explore_mode:
            self.hp.current_space[self.name] = list(self.options.keys())
            return self.explore()
        return self.handle_overrides_selections()

    def explore(self) -> Any:
        if self.name in self.hp.current_combination:
            selected_key = self.hp.current_combination[self.name]
        else:
            selected_key = list(self.options.keys())[0]
            self.hp.current_combination[self.name] = selected_key

        return selected_key


class MultiSelectCall(HPCall):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.all_combinations = None

    def execute(self, options: Union[Dict[str, Any], List[Any]]) -> List[Any]:
        processed_options = self.validate_and_process_options(options)
        self.options = processed_options

        if self.hp.explore_mode:
            if self.name in self.hp.current_combination:
                chosen_keys = self.hp.current_combination[self.name]
                return chosen_keys

            combinations = self._generate_all_combinations()
            self.hp.current_space[self.name] = combinations
            return self.explore(combinations)

        result = self.handle_overrides_selections()

        logger.info(f"MultiSelect call executed for {self.name}: {result}")
        return result

    def explore(self, combinations) -> List[Any]:
        selected_combination = combinations[0]
        self.hp.current_combination[self.name] = selected_combination
        return selected_combination

    def _generate_all_combinations(self) -> List[List[str]]:
        keys = list(self.options.keys())
        all_combinations = [[]]  # Include empty list for none selected
        for r in range(1, len(keys) + 1):
            all_combinations.extend(itertools.combinations(keys, r))
        return [list(combo) for combo in all_combinations]


class TextInputCall(HPCall):
    def execute(self, options: None = None) -> str:
        result = self.handle_overrides_selections()
        logger.info(f"TextInput call executed for {self.name}: {result}")
        return result

    def explore(self) -> str:
        return self.default if self.default is not None else ""


class NumberInputCall(HPCall):
    def execute(self, options: None = None) -> Union[int, float]:
        result = self.handle_overrides_selections()
        logger.info(f"NumberInput call executed for {self.name}: {result}")
        return result

    def explore(self) -> Union[int, float]:
        # TODO: handle no defaults in combinations
        return self.default

class IntInputCall(HPCall):
    def execute(self, options: None = None) -> int:
        result = self.handle_overrides_selections()
        logger.info(f"IntInput call executed for {self.name}: {result}")
        return result

    def explore(self) -> Union[int, float]:
        # TODO: handle no defaults in combinations
        return self.default

class MultiTextCall(HPCall):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.current_values = self.default or []

    def execute(self, options: None = None) -> List[str]:
        result = self.handle_overrides_selections()

        logger.info(f"MultiText call executed for {self.name}: {result}")
        return result

    def explore(self) -> List[str]:
        pass


class MultiNumberCall(HPCall):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.current_values = self.default or []

    def execute(self, options: None = None) -> List[Union[int, float]]:
        result = self.handle_overrides_selections()

        logger.info(f"MultiNumber call executed for {self.name}: {result}")
        return result

    def explore(self) -> List[Union[int, float]]:
        pass


class PropagateCall(HPCall):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def execute(
        self, config_func: Callable
    ) -> Dict[str, Any]:
        if self.hp.explore_mode:
            # TODO: make sure that this condition is valid and this doesn't need to be recalculated every time
            if self.name in self.hp.current_combination:
                selected_dct = self.hp.current_combination[self.name]
                return selected_dct

            combinations = config_func.get_combinations()
            defaults = config_func.get_defaults()
            self.hp.defaults[self.name] = defaults
            self.hp.current_space[self.name] = combinations
            return self.explore(combinations)

        original_name_prefix = self.hp.name_prefix
        self.hp.name_prefix = (
            self.name
            if original_name_prefix is None
            else f"{original_name_prefix}.{self.name}"
        )
        nested_config = self._prepare_nested_config()

        result = self._run_nested_config(config_func, nested_config)
        nested_snapshot = config_func.get_last_snapshot()
        self.hp.snapshot[self.name] = nested_snapshot
        self.hp.name_prefix = original_name_prefix
        return result

    def explore(self, combinations: List[Dict[str, Any]]) -> Dict[str, Any]:
        chosen_combination = combinations[0]
        self.hp.current_combination[self.name] = chosen_combination
        return chosen_combination

    def _prepare_nested_config(self) -> Dict[str, Any]:
        return {
            "selections": self.extract_nested_config(self.hp.selections, self.name),
            "overrides": self.extract_nested_config(self.hp.overrides, self.name),
            "final_vars": self.process_final_vars(self.hp.final_vars, self.name),
        }

    def _run_nested_config(
        self, config_func: Callable, nested_config: Dict[str, Any]
    ) -> Dict[str, Any]:
        logger.debug(f"Running nested configuration with: {nested_config}")
        return config_func(
            final_vars=nested_config["final_vars"],
            selections=nested_config["selections"],
            overrides=nested_config["overrides"],
        )

    @staticmethod
    def extract_nested_config(config: Dict[str, Any], prefix: str) -> Dict[str, Any]:
        result = collections.defaultdict(dict)
        prefix_dot = f"{prefix}."

        def process_item(key: str, value: Any, target: Dict[str, Any]):
            if key.startswith(prefix_dot):
                nested_key = key[len(prefix_dot) :]
                target[nested_key] = value
            elif key == prefix and isinstance(value, dict):
                target.update(value)

        for key, value in config.items():
            process_item(key, value, result)

        # Check for discrepancies
        flat_result = {k: v for k, v in PropagateCall._flatten_dict(result).items()}
        flat_dot_notation = {
            k[len(prefix_dot) :]: v
            for k, v in config.items()
            if k.startswith(prefix_dot)
        }

        # Check for keys with different values
        discrepancies = {
            k: (flat_result[k], flat_dot_notation[k])
            for k in set(flat_result) & set(flat_dot_notation)
            if flat_result[k] != flat_dot_notation[k]
        }

        if discrepancies:
            raise ValueError(
                f"Discrepancies found in nested configuration for '{prefix}': {discrepancies}"
            )

        return dict(result)

    @staticmethod
    def _flatten_dict(
        d: Dict[str, Any], parent_key: str = "", sep: str = "."
    ) -> Dict[str, Any]:
        items = []
        for k, v in d.items():
            new_key = f"{parent_key}{sep}{k}" if parent_key else k
            if isinstance(v, dict):
                items.extend(PropagateCall._flatten_dict(v, new_key, sep=sep).items())
            else:
                items.append((new_key, v))
        return dict(items)

    @staticmethod
    def process_final_vars(final_vars: List[str], name: str) -> List[str]:
        dot_notation_vars = [
            var[len(name) + 1 :] for var in final_vars if var.startswith(f"{name}.")
        ]
        dict_style_vars = list(
            PropagateCall.extract_nested_config(
                dict.fromkeys(final_vars, None), name
            ).keys()
        )
        dict_style_vars = [
            var for var in dict_style_vars if var not in dot_notation_vars
        ]  # remove dups
        return dot_notation_vars + dict_style_vars



import inspect
from typing import Callable, Union, overload

from .core import Hypster
from .hp import HP
from .utils import find_hp_function_body_and_name

HPFunc = Callable[[HP], None]


def create_hypster_instance(func: HPFunc, inject_names: bool) -> Hypster:
    """
    Create a Hypster instance from a configuration function.

    Args:
        func (HPFunc): The configuration function.
        inject_names (bool): Whether to inject names into the source code.

    Returns:
        Hypster: An instance of the Hypster class.

    Raises:
        ValueError: If no configuration function is found in the module.
    """
    source_code = inspect.getsource(func)
    result = find_hp_function_body_and_name(source_code)
    if result is None:
        raise ValueError("No configuration function found in the module")
    config_name, config_body = result
    namespace = {"HP": HP}
    return Hypster(config_name, config_body, namespace, inject_names=inject_names)


@overload
def config(func: HPFunc) -> Hypster: ...


@overload
def config(*, inject_names: bool = True) -> Callable[[HPFunc], Hypster]: ...


def config(
    func: Union[HPFunc, None] = None, *, inject_names: bool = True
) -> Union[Hypster, Callable[[HPFunc], Hypster]]:
    """
    Decorator to create a Hypster instance from a configuration function.

    This decorator can be used in two ways:
    1. As a simple decorator: @config
    2. With arguments: @config(inject_names=False)

    Args:
        func: Internal parameter to enable both decorator usage patterns.
              Users should not pass this argument directly.
        inject_names (bool, optional): Whether to automatically infer and inject
            parameter names into the source code. Defaults to True.

    Returns:
        Hypster: An instance of the Hypster class.

    Note:
        Although the return type is annotated as Union[Hypster, Callable[[HPFunc], Hypster]]
        for technical reasons, this function always returns a Hypster instance when used as
        a decorator.
    """

    def decorator(func: HPFunc) -> Hypster:
        return create_hypster_instance(func, inject_names)

    if func is None:
        return decorator
    return decorator(func)



# File: magic.py
import json
from pathlib import Path
from typing import Any, Dict, List

import ipywidgets as widgets
import pandas as pd
from IPython.core.magic import Magics, cell_magic, magics_class
from IPython.core.magic_arguments import argument, magic_arguments, parse_argstring
from IPython.display import HTML, clear_output, display

from .core import Hypster, find_hp_function_body_and_name
from .hp import HP
from .selection_handler import SelectionHandler


def custom_load(module_source: str, inject_names=True) -> Hypster:
    """
    Custom loader to parse the configuration module source code.

    Args:
        module_source (str): The source code of the configuration module.
        inject_names (bool): Whether to inject variable names into the namespace.

    Returns:
        Hypster: An instance of the Hypster configuration.
    """
    module_source = module_source.replace("@config", "").replace("@hypster.config", "")
    namespace = {"HP": HP}
    exec(module_source, namespace)
    result = find_hp_function_body_and_name(module_source)

    if result is None:
        raise ValueError("No configuration function found in the module")

    func_name, config_body = result
    func = namespace.get(func_name)

    if func is None:
        raise ValueError(f"Could not find the function {func_name} in the loaded module")

    return Hypster(func_name, config_body, namespace, inject_names)


class InteractiveHypster:
    """
    Interactive UI for Hypster configurations using ipywidgets.
    """

    def __init__(
        self,
        hp_config: Hypster,
        shell,
        selections: Dict[str, Any] = None,
        overrides: Dict[str, Any] = None,
        final_vars: List[str] = None,
        results_var: str = None,
        instantiate_on_first: bool = False,
        instantiate_on_change: bool = False,
        display_mode: str = "text",
    ):
        self.hp_config = hp_config
        self.shell = shell
        self.selection_handler = SelectionHandler(hp_config)
        self.selection_handler.initialize()
        self.overrides = overrides or {}
        self.final_vars = final_vars or []
        self.results_var = results_var
        self.instantiate_on_first = instantiate_on_first
        self.instantiate_on_change = instantiate_on_change
        self.widgets = {}
        self.output = widgets.Output()
        self.widgets_container = widgets.VBox([])
        self.selections = selections or {}
        self.display_mode = display_mode
        self.create_widgets()
        self.update_widgets_display()
        self.apply_selections()
        if self.instantiate_on_first:
            self.instantiate(None)

    def create_widgets(self):
        """
        Create widgets based on the current state of selections.
        """
        state = self.selection_handler.get_current_state()
        # Iterate in the order of selected_params to maintain widget order
        for param in state["selected_params"]:
            if param in state["current_options"]:
                options = state["current_options"][param]
                value = state["selected_params"][param]
                self.create_widget_for_param(param, options, value)

    def create_widget_for_param(self, param, options, value):
        """
        Create a widget for a given parameter based on its type.

        Args:
            param (str): The parameter name.
            options (Any): The options available for the parameter.
            value (Any): The current value of the parameter.
        """
        if param in self.overrides:
            value = self.overrides[param]
            disabled = True
        else:
            disabled = False

        if isinstance(options, dict):
            widget = self.create_dict_widget(param, options, value)
        elif isinstance(value, list):
            widget = self.create_list_widget(param, options, value)
        else:
            widget = self.create_singular_widget(param, options, value)

        widget.disabled = disabled
        self.widgets[param] = widget

    def create_dict_widget(self, param, options, value):
        """
        Create an accordion widget for dictionary-type parameters.

        Args:
            param (str): The parameter name.
            options (dict): Nested options for the parameter.
            value (dict): The current nested values for the parameter.

        Returns:
            widgets.Accordion: The accordion widget containing sub-widgets.
        """
        sub_widgets = []
        for sub_param, sub_options in options.items():
            sub_value = value.get(sub_param, None)
            sub_widget = self.create_widget_for_param(f"{param}.{sub_param}", sub_options, sub_value)
            sub_widgets.append(sub_widget)

        accordion = widgets.Accordion(children=sub_widgets)
        accordion.set_title(0, param)
        return accordion

    def create_list_widget(self, param, options, value):
        """
        Create a multi-select widget for list-type parameters.

        Args:
            param (str): The parameter name.
            options (set): The available options.
            value (list): The current selected values.

        Returns:
            widgets.SelectMultiple: The multi-select widget.
        """
        widget = widgets.SelectMultiple(
            options=sorted(list(options)),
            value=value,
            description=f"Select {param}",
            style={"description_width": "initial"},
        )
        # Observe changes to the 'value' property
        widget.observe(lambda change: self.on_change_list(param, change["new"]), names="value")
        return widget

    def on_change_list(self, param, new_value):
        """
        Handle changes in multi-select widgets.

        Args:
            param (str): The parameter name.
            new_value (tuple): The new selected values.
        """
        new_value = list(new_value)  # Convert from tuple to list
        # Update the parameter and refresh widgets
        self.selection_handler.update_param(param, new_value)
        self.update_widgets()
        # self.display_valid_combinations()

        if self.instantiate_on_change:
            self.instantiate(None)

    def create_singular_widget(self, param, options, value):
        """
        Create a dropdown widget for singular-type parameters.

        Args:
            param (str): The parameter name.
            options (set): The available options.
            value (Any): The current selected value.

        Returns:
            widgets.Dropdown: The dropdown widget.
        """
        widget = widgets.Dropdown(
            options=sorted(list(options)),
            value=value,
            description=f"Select {param}",
            style={"description_width": "initial"},
        )
        # Observe changes to the 'value' property
        widget.observe(lambda change: self.on_change(param, change["new"]), names="value")
        return widget

    def on_change(self, param, new_value):
        """
        Handle changes in singular widgets.

        Args:
            param (str): The parameter name.
            new_value (Any): The new selected value.
        """
        # Update the parameter and refresh widgets
        self.selection_handler.update_param(param, new_value)
        self.update_widgets()
        # self.display_valid_combinations()

        if self.instantiate_on_change:
            self.instantiate(None)

    def update_widgets(self):
        """
        Update existing widgets and create/remove widgets based on the current state.
        """
        state = self.selection_handler.get_current_state()
        current_params = set(state["selected_params"].keys())
        existing_params = set(self.widgets.keys())

        # Remove widgets for params that no longer exist
        for param in list(existing_params - current_params):
            del self.widgets[param]

        # Update or create widgets for current params
        for param in state["selected_params"]:
            if param in state["current_options"]:
                options = state["current_options"][param]
                value = state["selected_params"][param]
                if param in self.widgets:
                    self.update_widget(param, options, value)
                else:
                    self.create_widget_for_param(param, options, value)

        self.update_widgets_display()

    def update_widget(self, param, options, value):
        """
        Update an existing widget with new options and values.

        Args:
            param (str): The parameter name.
            options (Any): The updated options.
            value (Any): The updated value.
        """
        widget = self.widgets[param]
        if isinstance(options, dict):
            for sub_param, sub_options in options.items():
                full_param = f"{param}.{sub_param}"
                sub_value = value.get(sub_param, None)
                if full_param in self.widgets:
                    self.update_widget(full_param, sub_options, sub_value)
                else:
                    self.create_widget_for_param(full_param, sub_options, sub_value)
        elif isinstance(widget, widgets.SelectMultiple):
            widget.options = sorted(list(options))
            widget.value = [v for v in value if v in options]
        else:
            sorted_options = sorted(list(options))
            widget.options = sorted_options
            # Ensure the selected value is still valid
            widget.value = value if value in options else sorted_options[0]

    def update_widgets_display(self):
        """
        Refresh the widgets container to reflect the current set of widgets while maintaining order.
        """
        state = self.selection_handler.get_current_state()
        children = []
        # Iterate in the order of 'selected_params' to maintain widget order
        for param in state["selected_params"]:
            if param in self.widgets:
                children.append(self.widgets[param])
        self.widgets_container.children = children

    def display_valid_combinations(self):
        """
        Display the valid parameter combinations based on current selections.
        """
        with self.output:
            clear_output(wait=True)
            valid_combinations = self.selection_handler.filtered_combinations
            if valid_combinations:
                if self.display_mode == "dataframe":
                    self.display_as_dataframe(valid_combinations)
                elif self.display_mode == "json":
                    self.display_as_json(valid_combinations)
                else:
                    self.display_as_text(valid_combinations)

    def display_as_dataframe(self, valid_combinations):
        """
        Display valid combinations as a pandas DataFrame.
        """
        df = pd.DataFrame(valid_combinations).drop_duplicates()
        display(df)

    def display_as_json(self, valid_combinations):
        """
        Display valid combinations as formatted JSON.
        """
        json_output = json.dumps(valid_combinations, indent=2)
        display(HTML(f"<pre>{json_output}</pre>"))

    def display_as_text(self, valid_combinations):
        """
        Display valid combinations as plain text.
        """
        text_output = "\n".join([str(combo) for combo in valid_combinations])
        print(text_output)

    def instantiate(self, button):
        """
        Instantiate the configuration based on current selections.

        Args:
            button (widgets.Button): The button widget triggering the instantiation.
        """
        with self.output:
            clear_output(wait=True)
            state = self.selection_handler.get_current_state()
            results = self.hp_config(
                final_vars=self.final_vars, selections=state["selected_params"], overrides=self.overrides
            )
            if self.results_var:
                self.shell.user_ns[self.results_var] = results
            else:
                self.shell.user_ns.update(results)

    def display(self):
        """
        Display the interactive widgets and instantiate button in the notebook.
        """
        instantiate_button = widgets.Button(description="Instantiate")
        instantiate_button.on_click(self.instantiate)

        if self.instantiate_on_change:
            button_layout = widgets.Layout(display="none")
        else:
            button_layout = widgets.Layout()

        instantiate_button.layout = button_layout

        display(widgets.VBox([self.widgets_container, instantiate_button, self.output]))

    def apply_selections(self):
        """
        Apply user-provided selections iteratively.
        """
        state = self.selection_handler.get_current_state()
        for param in state["selected_params"]:
            if param in self.selections:
                self.update_param_with_selection(param, self.selections[param])

    def update_param_with_selection(self, param, value):
        """
        Update a parameter with a selected value and refresh the UI.
        """
        if param in self.widgets:
            widget = self.widgets[param]
            if isinstance(widget, widgets.SelectMultiple):
                widget.value = [v for v in value if v in widget.options]
            elif isinstance(widget, widgets.Dropdown):
                if value in widget.options:
                    widget.value = value
            elif isinstance(widget, widgets.Accordion):
                for i, sub_widget in enumerate(widget.children):
                    sub_param = f"{param}.{widget.get_title(i)}"
                    if sub_param in self.selections:
                        self.update_param_with_selection(sub_param, self.selections[sub_param])

            # Trigger the on_change event to update dependent parameters
            if isinstance(widget, widgets.SelectMultiple):
                self.on_change_list(param, widget.value)
            elif isinstance(widget, widgets.Dropdown):
                self.on_change(param, widget.value)


@magics_class
class HypsterMagics(Magics):
    """
    Custom IPython magic for Hypster configurations.
    """

    def __init__(self, shell):
        super().__init__(shell)
        self._first_run = True

    @magic_arguments()
    @argument("config_name", help="Name for the config module")
    @argument("-s", "--selections", help="Variable name containing selections dict")
    @argument("-o", "--overrides", help="Variable name containing overrides dict")
    @argument("-f", "--final_vars", help="Comma-separated list of final variables")
    @argument("-w", "--write_to_file", help="Write cell content to a file")
    @argument("-r", "--results", help="Variable name to store the results dictionary")
    @argument(
        "-i",
        "--instantiate",
        choices=["first", "change", "button", "first,change"],
        help="Instantiation behavior: 'first' (on first run), 'change' (on parameter change), "
        "'button' (manual only), or 'first,change' (on first run and parameter change)",
    )
    @cell_magic
    def hypster(self, line, cell):
        """
        Cell magic to initialize InteractiveHypster with given configuration.

        Usage:
            %%hypster config_name --results results_var -i first,change
            # Configuration code here
        """
        # Inject custom CSS to style the widgets
        css = """
        <style>
        .cell-output-ipywidget-background {
           background-color: transparent !important;
        }
        :root {
            --jp-widgets-color: var(--vscode-editor-foreground);
            --jp-widgets-font-size: var(--vscode-editor-font-size);
        }
        </style>
        """
        display(HTML(css))

        # Parse magic arguments
        args = parse_argstring(self.hypster, line)
        hp_config = custom_load(cell)

        # Retrieve selections and overrides from user namespace
        selections = self.shell.user_ns.get(args.selections, {}) if args.selections else {}
        overrides = self.shell.user_ns.get(args.overrides, {}) if args.overrides else {}
        final_vars = args.final_vars.split(",") if args.final_vars else None
        results_var = args.results

        # Determine instantiation behavior
        instantiate_options = args.instantiate.split(",") if args.instantiate else ["button"]
        instantiate_on_first = "first" in instantiate_options
        instantiate_on_change = "change" in instantiate_options

        # Initialize and display InteractiveHypster
        interactive_hypster = InteractiveHypster(
            hp_config,
            self.shell,
            selections=selections,  # Pass the selections here
            overrides=overrides,
            final_vars=final_vars,
            results_var=results_var,
            instantiate_on_first=instantiate_on_first,
            instantiate_on_change=instantiate_on_change,
        )
        interactive_hypster.display()

        # Store the configuration in the user namespace
        self.shell.user_ns[args.config_name] = hp_config

        # Optionally write the configuration to a file
        if args.write_to_file:
            Path(args.write_to_file).write_text(cell)

        return None


def load_ipython_extension(ipython):
    """
    Register the HypsterMagics with IPython.
    """
    ipython.register_magics(HypsterMagics)



import logging


class CustomFormatter(logging.Formatter):
    BLACK = "\033[0;30m"
    RED = "\033[0;31m"
    GREEN = "\033[0;32m"
    BROWN = "\033[0;33m"
    BLUE = "\033[0;34m"
    PURPLE = "\033[0;35m"
    CYAN = "\033[0;36m"
    LIGHT_GRAY = "\033[0;37m"
    DARK_GRAY = "\033[1;30m"
    LIGHT_RED = "\033[1;31m"
    LIGHT_GREEN = "\033[1;32m"
    YELLOW = "\033[1;33m"
    LIGHT_BLUE = "\033[1;34m"
    LIGHT_PURPLE = "\033[1;35m"
    LIGHT_CYAN = "\033[1;36m"
    LIGHT_WHITE = "\033[1;37m"
    BOLD = "\033[1m"
    FAINT = "\033[2m"
    ITALIC = "\033[3m"
    UNDERLINE = "\033[4m"
    BLINK = "\033[5m"
    NEGATIVE = "\033[7m"
    CROSSED = "\033[9m"
    END = "\033[0m"
    reset = "\x1b[0m"
    format = "%(message)s"

    FORMATS = {
        logging.DEBUG: CYAN + "%(levelname)s" + reset + " - " + format,
        logging.INFO: GREEN + "%(levelname)s" + reset + " - " + format,
        logging.WARNING: YELLOW + "%(levelname)s" + reset + " - " + format,
        logging.ERROR: RED + "%(levelname)s" + reset + " - " + format,
    }


# Logging Configuration Function
def configure_logging(level: int = logging.INFO) -> logging.Logger:
    # Get the root logger
    logger = logging.getLogger()

    # Remove all existing handlers
    if logger.hasHandlers():
        logger.handlers.clear()

    # Initialize the handler with the custom formatter
    handler = logging.StreamHandler()
    handler.setFormatter(CustomFormatter())

    # Set the handler for the logger
    logger.addHandler(handler)
    #logger.setLevel(level)
    logger.setLevel(logging.WARNING)

    return logger



import logging
import os
import textwrap
import types
from typing import Any, Dict, List, Optional, Tuple, Union

from .ast_analyzer import (
    collect_hp_calls,
    find_independent_select_calls,
    find_referenced_vars,
    inject_names_to_source_code,
)
from .hp import HP
from .utils import find_hp_function_body_and_name, remove_function_signature

# Correct logging configuration
# logging.basicConfig(level=logging.WARNING, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)

HypsterReturn = Union[Dict[str, Any], Tuple[Dict[str, Any], Dict[str, Any]]]


class Hypster:
    def __init__(self, name, source_code: str, namespace: Dict[str, Any], inject_names: bool = True):
        """
        Initialize a Hypster instance.

        Args:
            name (str): The name of the Hypster instance.
            source_code (str): The source code to be executed.
            namespace (Dict[str, Any]): The namespace for execution.
            inject_names (bool, optional): Whether to inject names into the source code. Defaults to True.
        """
        self.name = name
        self.source_code = source_code
        self.namespace = namespace
        self.hp_calls = collect_hp_calls(self.source_code)

        self.modified_source = (
            inject_names_to_source_code(self.source_code, self.hp_calls) if inject_names else self.source_code
        )

        self.referenced_vars = find_referenced_vars(self.source_code)
        self.independent_select_calls = find_independent_select_calls(self.referenced_vars, self.hp_calls)
        self.combinations = []
        self.defaults = {}
        self.snapshot_history = []

    def __call__(
        self,
        final_vars: Optional[List[str]] = None,
        selections: Optional[Dict[str, Any]] = None,
        overrides: Optional[Dict[str, Any]] = None,
    ) -> HypsterReturn:
        """
        Execute the Hypster instance with given parameters.

        Args:
            final_vars (Optional[List[str]], optional): List of variables to include in the final result.
            selections (Optional[Dict[str, Any]], optional): Predefined selections for hyperparameters.
            overrides (Optional[Dict[str, Any]], optional): Overrides for hyperparameters.

        Returns:
            Dict[str, Any]: The execution result.
        """
        hp = HP(final_vars or [], selections or {}, overrides or {})
        result = self._execute_function(hp, self.modified_source)
        self.snapshot_history.append(hp.snapshot)
        return result

    def _execute_function(self, hp: HP, modified_source: str) -> Dict[str, Any]:
        """
        Execute the modified source code with the given HP instance.

        Args:
            hp (HP): The HP instance for selections & overrides management.
            modified_source (str): The modified source code to execute.

        Returns:
            Dict[str, Any]: The instantiated config.
        """

        body_wo_signature = remove_function_signature(modified_source)
        function_body = textwrap.dedent(body_wo_signature)

        # Create a new namespace with the original namespace and add the 'hp' object to it
        exec_namespace = self.namespace.copy()
        exec_namespace["hp"] = hp

        # Execute the modified function body in this namespace
        exec(function_body, exec_namespace)

        # Process and filter the results
        return self._process_results(exec_namespace, hp.final_vars)

    def _process_results(self, namespace: Dict[str, Any], final_vars: List[str]) -> Dict[str, Any]:
        """
        Process and filter the execution results.

        Args:
            namespace (Dict[str, Any]): The namespace after execution.
            final_vars (List[str]): List of variables to include in the final result.

        Returns:
            Dict[str, Any]: The processed and filtered results.

        Raises:
            ValueError: If any variable in final_vars does not exist in the configuration.
        """

        filtered_locals = {
            k: v
            for k, v in namespace.items()
            if k != "hp" and not k.startswith("__") and not isinstance(v, (types.ModuleType, types.FunctionType, type))
        }

        if not final_vars:
            final_result = {k: v for k, v in filtered_locals.items() if not k.startswith("_")}
        else:
            non_existent_vars = set(final_vars) - set(filtered_locals.keys())
            if non_existent_vars:
                raise ValueError(
                    "The following variables specified in final_vars "
                    "do not exist in the configuration: "
                    f"{', '.join(non_existent_vars)}"
                )
            final_result = {k: filtered_locals[k] for k in final_vars}

        logger.debug("Captured locals: %s", filtered_locals)
        logger.debug("Final result after filtering: %s", final_result)

        return final_result

    def save(self, path: Optional[str] = None):
        """
        Save the current object to a file.

        Parameters:
        path (Optional[str]): The file path where the object should be saved.
                              If None, the object will be saved with its name as the filename.

        Returns:
        None
        """
        if path is None:
            path = f"{self.name}.py"
        save(self, path)

    def get_combinations(self):
        if self.combinations:
            return self.combinations

        hp = HP([], {}, {}, explore_mode=True)
        combinations = []

        while True:
            self._execute_function(hp, self.modified_source)
            combinations.append(hp.current_combination.copy())
            if not hp.get_next_combination(combinations):
                break
        self.combinations = combinations
        self.defaults = hp.defaults  # TODO: improve this
        return combinations

    def get_defaults(self):
        if not self.combinations:
            self.get_combinations()
        return self.defaults

    def get_snapshot_history(self):
        return self.snapshot_history

    def get_last_snapshot(self):
        if not self.snapshot_history:
            return {}
        return self.snapshot_history[-1]


def save(hypster_instance: Hypster, path: Optional[str] = None):
    """
    Save the configuration of a Hypster instance to a file.
    This function extracts the configuration function from the provided
    Hypster instance's source code and writes it to a specified file. If no
    path is provided, the configuration is saved to a file named after the
    Hypster instance's name with a .py extension.
    Args:
        hypster_instance (Hypster): The Hypster instance whose configuration
            is to be saved.
        path (Optional[str]): The file path where the configuration should be
            saved. If None, the configuration is saved to a file named after
            the Hypster instance's name.
    Raises:
        ValueError: If the provided object is not a Hypster instance or if no
            configuration function is found in the module.
    Returns:
        None
    """
    if not isinstance(hypster_instance, Hypster):
        raise ValueError("The provided object is not a Hypster instance")

    if path is None:
        path = f"{hypster_instance.name}.py"

    result = find_hp_function_body_and_name(hypster_instance.source_code)

    if result is None:
        raise ValueError("No configuration function found in the module")

    func_name, hp_func_source = result

    modified_source = "from hypster import HP\n\n" + hp_func_source
    directory = os.path.dirname(path)
    if directory and not os.path.exists(directory):
        os.makedirs(directory, exist_ok=True)
    with open(path, "w") as f:
        f.write(modified_source)

    logger.info("Configuration saved to %s", path)


def load(path: str, inject_names=True) -> Hypster:
    """
    Loads a Python module from the specified file path, executes it, and retrieves a configuration function.
    Args:
        path (str): The file path to the Python module to be loaded.
        inject_names (bool, optional): If True, injects names into the namespace. Defaults to True.

    Returns:
        Hypster: An instance of the Hypster class containing the configuration function and its context.

    Raises:
        ValueError: If no configuration function is found in the module or
                    if the function cannot be retrieved from the namespace.
    """
    with open(path, "r") as f:
        module_source = f.read()

    # Create a namespace with HP already imported
    namespace = {"HP": HP}

    # Execute the entire module
    exec(module_source, namespace)

    # Find the configuration function
    result = find_hp_function_body_and_name(module_source)

    if result is None:
        raise ValueError("No configuration function found in the module")

    # Unpack the function name and body from the result
    func_name, config_body = result

    # Retrieve the function object from the namespace
    func = namespace.get(func_name)

    if func is None:
        raise ValueError(f"Could not find the function {func_name} in the loaded module")

    return Hypster(func_name, config_body, namespace, inject_names)



import logging
from collections import OrderedDict
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from .hp_calls import (
    MultiNumberCall,
    MultiSelectCall,
    MultiTextCall,  # New imports
    NumberInputCall,
    IntInputCall,
    PropagateCall,
    SelectCall,
    TextInputCall,
)

#logging.basicConfig(level=logging.WARNING, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


class HP:
    def __init__(
        self, final_vars: List[str], selections: Dict[str, Any], overrides: Dict[str, Any], explore_mode: bool = False
    ):
        self.final_vars = final_vars
        self.selections = selections
        self.overrides = overrides
        self.name_prefix = None
        self.explore_mode = explore_mode
        self.current_space = OrderedDict()  # TODO: consider turning these 3 items to a dataclass
        self.current_combination = OrderedDict()
        self.defaults = {}
        self.snapshot = {}
        logger.info(f"Initialized HP with explore_mode: {explore_mode}")

    def select(self, options: Union[Dict[str, Any], List[Any]], *, name: Optional[str] = None, default: Any = None):
        select_call = SelectCall(self, name=name, default=default)
        logger.debug(f"Added SelectCall: {name}")
        return select_call.execute(options)

    # TODO: add "min_items" to multi_select
    def multi_select(
        self, options: Union[Dict[str, Any], List[Any]], *, name: Optional[str] = None, default: Any = None
    ) -> List[Any]:
        multi_select_call = MultiSelectCall(self, name=name, default=default)
        logger.debug(f"Added MultiSelectCall: {name}")
        return multi_select_call.execute(options)

    def text_input(self, default: str, *, name: Optional[str] = None) -> str:
        text_input_call = TextInputCall(self, name=name, default=default)
        logger.debug(f"Added TextInputCall: {name}")
        return text_input_call.execute()

    def number_input(self, default: Union[int, float], *, name: Optional[str] = None) -> Union[int, float]:
        number_input_call = NumberInputCall(self, name=name, default=default)
        logger.debug(f"Added NumberInputCall: {name}")
        return number_input_call.execute()

    def int_input(self, default: int, *, name: Optional[str] = None) -> int:
        int_input_call = IntInputCall(self, name=name, default=default)
        logger.debug(f"Added IntInputCall: {name}")
        return int_input_call.execute()

    def multi_text(self, default: List[str] = [], *, name: Optional[str] = None) -> List[str]:
        multi_text_call = MultiTextCall(self, name=name, default=default)
        logger.debug(f"Added MultiTextCall: {name}")
        return multi_text_call.execute()

    def multi_number(
        self, default: List[Union[int, float]] = [], *, name: Optional[str] = None
    ) -> List[Union[int, float]]:
        multi_number_call = MultiNumberCall(self, name=name, default=default)
        logger.debug(f"Added MultiNumberCall: {name}")
        return multi_number_call.execute()

    def propagate(
        self, config_func: Callable, *, name: Optional[str] = None
    ) -> Dict[str, Any]:
        propagate_call = PropagateCall(self, name=name)
        logger.debug(f"Added PropagateCall: {name}")
        return propagate_call.execute(config_func)

    def get_next_combination(self, explored_combinations: List[Dict[str, Any]]) -> bool:
        for param in reversed(self.current_combination.keys()):
            options = self.current_space[param]
            current_value = self.current_combination[param]

            next_option = self._get_next_option(param, options, current_value, explored_combinations)
            if next_option is not None:
                self.current_combination[param] = next_option
                self._reset_subsequent_params(param)
                return True

        return False

    def _flatten_combination(self, combination: OrderedDict) -> OrderedDict:
        flattened = OrderedDict()
        for key, value in combination.items():
            if isinstance(value, dict):
                for sub_key, sub_value in self._flatten_combination(value).items():
                    flattened[f"{key}.{sub_key}"] = sub_value
            else:
                flattened[key] = value
        return flattened

    def _get_next_option(
        self, param: str, options: List[Any], current_value: Any, explored_combinations: List[Dict[str, Any]]
    ) -> Optional[Any]:
        param_index = list(self.current_combination.keys()).index(param)
        previous_selections = {
            k: v
            for k, v in self.current_combination.items()
            if list(self.current_combination.keys()).index(k) < param_index
        }
        used_options = set(
            _hashable_value(comb[param])
            for comb in explored_combinations
            if all(_values_equal(comb.get(k), v) for k, v in previous_selections.items())
        )

        for option in options:
            if _hashable_value(option) not in used_options:
                return option

        return None

    def _reset_subsequent_params(self, param: str):
        reset = False
        for name in list(self.current_combination.keys()):
            if reset:
                self.current_combination.pop(name)
            if name == param:
                reset = True


class InvalidSelectionError(Exception):
    pass


def _hashable_value(value: Any) -> Tuple:
    if isinstance(value, dict):
        return tuple(sorted((k, _hashable_value(v)) for k, v in value.items()))
    elif isinstance(value, list):
        return tuple(_hashable_value(item) for item in value)
    return value


def _values_equal(a: Any, b: Any) -> bool:
    if isinstance(a, dict) and isinstance(b, dict):
        return all(_values_equal(a.get(k), b.get(k)) for k in set(a.keys()) | set(b.keys()))
    elif isinstance(a, list) and isinstance(b, list):
        return len(a) == len(b) and all(_values_equal(x, y) for x, y in zip(a, b))
    return a == b
